## Copyright (C) 2014 Bitergia
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## This file is a part of GrimoireLib
##  (an Python library for the MetricsGrimoire and vizGrimoire systems)
##
##
## Authors:
##   Alvaro del Castillo <acs@bitergia.com>

""" Metrics for the issue tracking system """

from .metrics import Metrics

class ITS():
    name = "its"

    @classmethod
    def get_section_metrics(cls):
        return {
            "overview" : {
                "activity_metrics": [Closed, Opened],
                "author_metrics": [],
                "bmi_metrics": [BMI],
                "time_to_close_metrics": [DaysToCloseMedian],
                "projects_metrics": [Projects]
            },
            "com_channels": {
                "activity_metrics": [],
                "author_metrics": []
            },
            "project_activity": {
                "metrics": [Opened, Closed]
            },
            "project_community": {
                "author_metrics": [],
                "people_top_metrics": [],
                "orgs_top_metrics": [],
            },
            "project_process": {
                "bmi_metrics": [BMI],
                "time_to_close_metrics": [DaysToCloseAverage, DaysToCloseMedian],
                "time_to_close_title": "Days to close (median and average)",
                "time_to_close_review_metrics": [],
                "time_to_close_review_title": "",
                "patchsets_metrics": []
            }
        }



class ITSMetrics(Metrics):
    ds = ITS


class Opened(ITSMetrics):
    """ Tickets Opened metric class for issue tracking systems """

    id = "opened"
    name = "Opened tickets"
    desc = "Number of opened tickets"
    FIELD_COUNT="id"
    FIELD_NAME="url"


class Openers(ITSMetrics):
    """ Tickets Openers metric class for issue tracking systems """

    id = "openers"
    name = "Ticket submitters"
    desc = "Number of persons submitting new tickets"


class Closed(ITSMetrics):
    """ Tickets Closed metric class for issue tracking systems
       "state": "closed" is the filter to be used in GitHub
    """
    id = "closed"
    name = "Closed tickets"
    desc = "Number of closed tickets"
    filters = {"state":"closed"}
    FIELD_COUNT="id"
    FIELD_NAME="url"


class DaysToCloseMedian(ITSMetrics):
    id = "days_to_close_ticket_median"
    name = "Days to close tickets (median)"
    desc = "Number of days needed to close a ticket (median)"
    FIELD_COUNT = 'time_to_close_days'
    AGG_TYPE = 'median'
    filters = {"state":"closed"}

    def get_agg(self):
        agg = super(DaysToCloseMedian, self).get_agg()
        if agg == None:
            agg = 0  # None is because NaN in ES. Let's convert to 0
        return agg


class DaysToCloseAverage(ITSMetrics):
    id = "days_to_close_ticket_avg"
    name = "Days to close tickets (average)"
    desc = "Number of days needed to close a ticket (average)"
    FIELD_COUNT = 'time_to_close_days'
    AGG_TYPE = 'average'
    filters = {"state":"closed"}


class Closers(ITSMetrics):
    """ Tickets Closers metric class for issue tracking systems """
    id = "closers"
    name = "Tickets closers"
    desc = "Number of persons closing tickets"


class BMI(ITSMetrics):
    """ The Backlog Management Index measures efficiency dealing with tickets

        This is based on the book "Metrics and Models in Software Quality
        Engineering. Chapter 4.3.1. By Stephen H. Kan.

        BMI is calculated as the number of closed tickets out of the opened
        tickets in a given interval. This metric aims at having an overview of
        how the community deals with tickets. Continuous values under 1
        (or 100 if this is calculated as a percentage) shows low peformance
        given that the community leaves a lot of opened tickets. Continuous
        values close to 1 or over 1 shows a better performance. This would
        indicate that most of the tickets are being closed.
    """

    id = "bmi_tickets"
    name = "Backlog Management Index"
    desc = "Number of tickets closed out of the opened ones in a given interval"

    def __get_metrics(self):
        """ Each metric must have its own filters copy to modify it freely"""
        esfilters_closed = None
        esfilters_opened = None
        if self.esfilters:
            esfilters_closed = self.esfilters.copy()
            esfilters_opened = self.esfilters.copy()

        closed = Closed(self.es_url, self.es_index,
                        start=self.start, end=self.end,
                        esfilters=esfilters_closed, interval=self.interval)
        # For BMI we need when the ticket was closed
        closed.FIELD_DATE = "closed_at"
        opened = Opened(self.es_url, self.es_index,
                        start=self.start, end=self.end,
                        esfilters=esfilters_opened, interval=self.interval)

        return (closed, opened)

    def get_agg(self):
        (closed, opened) = self.__get_metrics()
        closed_agg = closed.get_agg()
        opened_agg = opened.get_agg()

        if opened_agg == 0:
            bmi = 1  # if no submitted prs, bmi is at 100%
        else:
            bmi = closed_agg/opened_agg

        return bmi

    def get_ts(self):
        bmi = {}
        (closed, opened) = self.__get_metrics()
        closed_ts = closed.get_ts()
        opened_ts = opened.get_ts()

        bmi['date'] = closed_ts['date']
        bmi['unixtime'] = closed_ts['unixtime']
        bmi['value'] = []
        for i in range(0, len(opened_ts['value'])):
            if opened_ts['value'][i] == 0:
                bmi['value'].append(0)
            else:
                bmi['value'].append(closed_ts['value'][i] / opened_ts['value'][i])

        return bmi


class Projects(ITSMetrics):
    """ Projects metric class for issue tracking systems """
    id = "projects"
    name = "Projects"
    desc = "Number of distinct projects active in the ticketing system"
    FIELD_NAME = 'project' # field used to list projects



## Copyright (C) 2014 Bitergia
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## This file is a part of GrimoireLib
##  (an Python library for the MetricsGrimoire and vizGrimoire systems)
##
##
## Authors:
##   Alvaro del Castillo <acs@bitergia.com>
##   Daniel Izquierdo <dizquierdo@bitergia.com>

from . import its
from .metrics import Metrics


class Jira(its.ITS):
    name = "jira_issues"

    @classmethod
    def get_section_metrics(cls):
        return {
            "overview" : {
                "activity_metrics": [Closed, Opened],
                "author_metrics": [],
                "bmi_metrics": [BMI],
                "time_to_close_metrics": [DaysToCloseMedian],
                "projects_metrics": [Projects]
            },
            "com_channels": {
                "activity_metrics": [],
                "author_metrics": []
            },
            "project_activity": {
                "metrics": [Opened, Closed]
            },
            "project_community": {
                "author_metrics": [],
                "people_top_metrics": [],
                "orgs_top_metrics": [],
            },
            "project_process": {
                "bmi_metrics": [BMI],
                "time_to_close_metrics": [DaysToCloseAverage, DaysToCloseMedian],
                "time_to_close_title": "Days to close (median and average)",
                "time_to_close_review_metrics": [],
                "time_to_close_review_title": "",
                "patchsets_metrics": []
            }
        }


class JiraMetrics(Metrics):
    # TODO: All JiraMetrics metrics should inherit from this class
    # but they are doing it from ITSMetrics directly. Change the design.
    ds = Jira

class Opened(JiraMetrics):
    """ Tickets Opened metric class for issue tracking systems """

    id = "opened"
    name = "Opened tickets"
    desc = "Number of opened tickets"
    FIELD_COUNT="key"
    FIELD_NAME="url"


class Openers(its.Openers):
    ds = Jira

class Closed(JiraMetrics):
    """ Tickets Closed metric class for issue tracking systems
       "status":"Closed"  "status":"Resolved"  "status":"Done"
    """
    id = "closed"
    name = "Closed tickets"
    desc = "Number of closed tickets"
    # filters = {"status":["Closed", "Resolved", "Done"]}
    filters = {"*status": "Open"}
    FIELD_COUNT="key"
    FIELD_NAME="url"


class DaysToCloseMedian(its.DaysToCloseMedian):
    ds = Jira

class DaysToCloseAverage(its.DaysToCloseAverage):
    ds = Jira

class Closers(its.Closers):
    ds = Jira

class BMI(its.BMI):
    ds = Jira

class Projects(its.Projects):
    ds = Jira



## Copyright (C) 2014 Bitergia
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## This file is a part of GrimoireLib
##  (an Python library for the MetricsGrimoire and vizGrimoire systems)
##
##
## Authors:
##   Alvaro del Castillo <acs@bitergia.com>
##   Daniel Izquierdo <dizquierdo@bitergia.com>

from .metrics import Metrics

class GitHubPRs():
    name = "github_prs"

    @classmethod
    def get_section_metrics(cls):
        # Those metrics are only for Pull Requests
        # github issues is covered as ITS
        return {
            "overview" : {
                "activity_metrics": [ClosedPR, SubmittedPR],
                "author_metrics": [],
                "bmi_metrics": [BMIPR],
                "time_to_close_metrics": [DaysToClosePRMedian],
                "projects_metrics": [Projects]
            },
            "com_channels": {
                "activity_metrics": [],
                "author_metrics": []
            },
            "project_activity": {
                "metrics": [SubmittedPR, ClosedPR]
            },
            "project_community": {
                "author_metrics": [],
                "people_top_metrics": [],
                "orgs_top_metrics": [],
            },
            "project_process": {
                "bmi_metrics": [BMIPR],
                "time_to_close_metrics": [],
                "time_to_close_title": "Days to close (median and average)",
                "time_to_close_review_metrics": [DaysToClosePRAverage, DaysToClosePRMedian],
                "time_to_close_review_title": "Days to close review (median and average)",
                "patchsets_metrics": []
            }
        }



class GitHubPRsMetrics(Metrics):
    ds = GitHubPRs


class SubmittedPR(GitHubPRsMetrics):
    id = "submitted"
    name = "Submitted reviews"
    desc = "Number of submitted code review processes"
    FIELD_NAME = 'id'
    FIELD_COUNT = 'id'
    filters = {"pull_request":"true"}


class ClosedPR(GitHubPRsMetrics):
    id = "closed"
    name = "Closed reviews"
    desc = "Number of closed review processes (merged or abandoned)"
    FIELD_NAME = 'id'
    FIELD_COUNT = 'id'
    filters = {"pull_request":"true", "state":"closed"}


class DaysToClosePRMedian(GitHubPRsMetrics):
    id = "days_to_close_pr_median"
    name = "Days to close reviews (median)"
    desc = "Number of days needed to close a review (median)"
    FIELD_COUNT = 'time_to_close_days'
    AGG_TYPE = 'median'
    filters = {"pull_request":"true", "state":"closed"}

    def get_agg(self):
        agg = super(type(self), self).get_agg()
        if agg == None:
            agg = 0  # None is because NaN in ES. Let's convert to 0
        return agg

class DaysToClosePRAverage(GitHubPRsMetrics):
    id = "days_to_close_pr_avg"
    name = "Days to close reviews (avg)"
    desc = "Number of days needed to close a review (average)"
    FIELD_COUNT = 'time_to_close_days'
    AGG_TYPE = 'average'
    filters = {"pull_request":"true", "state":"closed"}


class Projects(GitHubPRsMetrics):
    """ Projects in the review code management system """

    id = "projects"
    name = "Projects"
    desc = "Projects in the review code management system"
    FIELD_NAME = 'project' # field used to list projects


class BMIPR(GitHubPRsMetrics):
    """This class calculates the efficiency closing reviews

    It is calculated as the number of closed issues out of the total number
    of opened ones in a period.
    """

    id = "bmipr"
    name = "BMI Pull Requests"
    desc = "Efficiency reviewing: (closed prs)/(submitted prs)"

    def __get_metrics(self):
        """ Each metric must have its own filters copy to modify it freely"""
        esfilters_close = None
        esfilters_submit = None
        if self.esfilters:
            esfilters_close = self.esfilters.copy()
            esfilters_submit = self.esfilters.copy()

        closed = ClosedPR(self.es_url, self.es_index,
                          start=self.start, end=self.end,
                          esfilters=esfilters_close, interval=self.interval)
        # For BMI we need when the ticket was closed
        closed.DATE_FIELD = 'updated_at'
        submitted = SubmittedPR(self.es_url, self.es_index,
                                start=self.start, end=self.end,
                                esfilters=esfilters_submit,
                                interval=self.interval)

        return (closed, submitted)

    def get_agg(self):
        (closed, submitted) = self.__get_metrics()
        closed_agg = closed.get_agg()
        submitted_agg = submitted.get_agg()

        if submitted_agg == 0:
            bmi = 1  # if no submitted prs, bmi is at 100%
        else:
            bmi = closed_agg/submitted_agg

        return bmi


    def get_ts(self):
        bmi = {}
        (closed, submitted) = self.__get_metrics()
        closed_ts = closed.get_ts()
        submitted_ts = submitted.get_ts()

        bmi['date'] = closed_ts['date']
        bmi['unixtime'] = closed_ts['unixtime']
        bmi['value'] = []
        for i in range(0, len(submitted_ts['value'])):
            if submitted_ts['value'][i] == 0:
                bmi['value'].append(0)
            else:
                bmi['value'].append(closed_ts['value'][i] / submitted_ts['value'][i])

        return bmi


class Reviewers(GitHubPRsMetrics):
    """ People assigned to pull requests """
    id = "reviewers"
    name = "Reviewers"
    desc = "Number of persons reviewing code review activities"


class Closers(GitHubPRsMetrics):
    id = "closers"
    name = "Closers"
    desc = "Number of persons closing code review activities"


# Pretty similar to ITS openers
class Submitters(GitHubPRsMetrics):
    id = "submitters"
    name = "Submitters"
    desc = "Number of persons submitting code review processes"



## Copyright (C) 2014 Bitergia
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## This file is a part of GrimoireLib
##  (an Python library for the MetricsGrimoire and vizGrimoire systems)
##
##
## Authors:
##   Daniel Izquierdo-Cortazar <dizquierdo@bitergia.com>
##   Alvaro del Castillo <acs@bitergia.com>
##

from .metrics import Metrics

class MLS():
    name = "mailing lists"

    @classmethod
    def get_section_metrics(cls):
        return {
            "overview" : {
                "activity_metrics": [EmailsSent],
                "author_metrics": [],
                "bmi_metrics": [],
                "time_to_close_metrics": [],
                "projects_metrics": [Projects]
            },
            "com_channels": {
                "activity_metrics": [EmailsSent],
                "author_metrics": [EmailsSenders]
            },
            "project_activity": {
                # TODO: EmailsSenders is not activity but we need two metrics here
                "metrics": [EmailsSent, EmailsSenders]
            },
            "project_community": {
                "author_metrics": [],
                "people_top_metrics": [],
                "orgs_top_metrics": [],
            },
            "project_process": {
                "bmi_metrics": [],
                "time_to_close_metrics": [],
                "time_to_close_title": "",
                "time_to_close_review_metrics": [],
                "time_to_close_review_title": "",
                "patchsets_metrics": []
            }
        }


class MLSMetrics(Metrics):
    ds = MLS


class EmailsSent(MLSMetrics):
    """ Emails metric class for mailing lists analysis """

    id = "emails_sent"
    name = "Sent Emails"
    desc = "Emails sent to mailing lists"
    FIELD_COUNT= "Message-ID"
    FIELD_NAME = "Message-ID"


class EmailsSenders(MLSMetrics):
    """ Emails Senders class for mailing list analysis """

    id = "emails_senders"
    name = "Email Senders"
    desc = "People sending emails"

    FIELD_COUNT = 'author_uuid' # field used to count Authors
    FIELD_NAME = 'author_name' # field used to count Authors


class Projects(MLSMetrics):
    """ Projects in the mailing lists """

    id = "projects"
    name = "Projects"
    desc = "Projects in the mailing lists"
    FIELD_NAME = 'project' # field used to list projects



## Copyright (C) 2017 Bitergia
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
##
## Authors:
##   Alvaro del Castillo <acs@bitergia.com>
##

from .metrics import Metrics

class Stackexchange():
    name = "stackexchange"

    @classmethod
    def get_section_metrics(cls):
        return {
            "overview" : {
                "activity_metrics": [QuestionsSent],
                "author_metrics": [],
                "bmi_metrics": [],
                "time_to_close_metrics": [],
                "projects_metrics": [Projects]
            },
            "com_channels": {
                "activity_metrics": [QuestionsSent],
                "author_metrics": [QuestionsSenders]
            },
            "project_activity": {
                # TODO: QuestionsSenders is not activity but we need two metrics here
                "metrics": [QuestionsSent, QuestionsSenders]
            },
            "project_community": {
                "author_metrics": [],
                "people_top_metrics": [],
                "orgs_top_metrics": [],
            },
            "project_process": {
                "bmi_metrics": [],
                "time_to_close_metrics": [],
                "time_to_close_title": "",
                "time_to_close_review_metrics": [],
                "time_to_close_review_title": "",
                "patchsets_metrics": []
            }
        }


class StackexchangeMetrics(Metrics):
    ds = Stackexchange


class QuestionsSent(StackexchangeMetrics):
    """ Questions metric class for stackexchange analysis """

    id = "questions_sent"
    name = "Sent Questions"
    desc = "Questions sent to stackexchange site"
    FIELD_COUNT = "uuid"  # field used to count questions
    FIELD_NAME = "title"  # field used to list qyestions
    filters = {"is_stackexchange_question":"1"}


class QuestionsSenders(StackexchangeMetrics):
    """ Questions Senders class for stackexchange analysis """

    id = "questions_senders"
    name = "Questions Senders"
    desc = "People sending questions"

    FIELD_COUNT = 'author_uuid' # field used to count Authors
    FIELD_NAME = 'author_name' # field used to count Authors
    filters = {"is_stackexchange_question":"1"}


class Projects(StackexchangeMetrics):
    """ Projects in the stackexchange """

    id = "projects"
    name = "Projects"
    desc = "Projects in stackexchange"
    FIELD_NAME = 'project' # field used to list projects



## Copyright (C) 2016 Bitergia
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
##
## Authors:
##   Alvaro del Castillo <acs@bitergia.com>
##   Daniel Izquierdo <dizquierdo@bitergia.com>

""" Metrics for Gerrit review system """

import logging

from .metrics import Metrics

class Gerrit():
    name = "gerrit"

    @classmethod
    def get_section_metrics(cls):
        return {
            "overview" : {
                "activity_metrics": [Closed, Submitted],
                "author_metrics": None,
                "bmi_metrics": [BMI],
                "time_to_close_metrics": [DaysToMergeMedian],
                "projects_metrics": [Projects],
            },
            "com_channels": {
                "activity_metrics": [],
                "author_metrics": []
            },
            "project_activity": {
                "metrics": [Submitted, Closed]
            },
            "project_community": {
                "author_metrics": [],
                "people_top_metrics": [],
                "orgs_top_metrics": [],
            },
            "project_process": {
                "bmi_metrics": [BMI],
                "time_to_close_metrics": [],
                "time_to_close_title": "",
                "time_to_close_review_metrics": [DaysToMergeAverage, DaysToMergeMedian],
                "time_to_close_review_title": "Days to close review (median and average)",
                "patchsets_metrics": [PatchsetsMedian, PatchsetsAverage],
                "patchsets_title": "Number of patchsets per review (median and average)"
            }
        }


class GerritMetrics(Metrics):
    ds = Gerrit


class Submitted(GerritMetrics):
    id = "submitted"
    name = "Submitted reviews"
    desc = "Number of submitted code review processes"
    FIELD_NAME = 'url'
    FIELD_COUNT = 'url'


class Merged(GerritMetrics):
    id = "merged"
    name = "Merged reviews"
    desc = "Number of merged review processes"
    FIELD_NAME = 'url'
    FIELD_COUNT = 'url'
    filters = {"status": "MERGED"}

class Abandoned(GerritMetrics):
    id = "abandoned"
    name = "Abandoned reviews"
    desc = "Number of abandoned review processes"
    FIELD_NAME = 'url'
    FIELD_COUNT = 'url'
    filters = {"status": "ABANDONED"}

class Closed(GerritMetrics):
    id = "closed"
    name = "Closed reviews"
    desc = "Number of closed review processes (merged or abandoned)"

    def __get_metrics(self):
        """ Each metric must have its own filters copy to modify it freely"""
        esfilters_merge = None
        esfilters_abandon = None
        if self.esfilters:
            esfilters_merge = self.esfilters.copy()
            esfilters_abandon = self.esfilters.copy()

        merged = Merged(self.es_url, self.es_index,
                              start=self.start, end=self.end,
                              esfilters=esfilters_merge, interval=self.interval)

        abandoned = Abandoned(self.es_url, self.es_index,
                                    start=self.start, end=self.end,
                                    esfilters=esfilters_abandon, interval=self.interval)
        return (merged, abandoned)

    def get_agg(self):
        (merged, abandoned) = self.__get_metrics()
        merged_agg = merged.get_agg()
        abandoned_agg = abandoned.get_agg()
        agg = merged_agg + abandoned_agg
        return agg

    def get_ts(self):
        closed = {}
        (merged, abandoned) = self.__get_metrics()
        merged_ts = merged.get_ts()
        abandoned_ts = abandoned.get_ts()

        closed['date'] = merged_ts['date']
        closed['unixtime'] = merged_ts['unixtime']
        closed['value'] = []
        for i in range(0, len(merged_ts['value'])):
            closed['value'].append(merged_ts['value'][i] + abandoned_ts['value'][i])
        return closed


class BMI(GerritMetrics):
    """This class calculates the efficiency closing reviews

    This class is based on the Backlog Management Index that in issues, it is
    calculated as the number of merged issues out of the total number of opened
    ones in a period. (The other way around also provides an interesting view).

    In terms of the code review system, this values is measured as the number
    of merged+abandoned reviews out of the total number of submitted ones.
    """

    id = "bmi_reviews"
    name = "BMI Gerrit"
    desc = "Efficiency reviewing: (merged+abandoned reviews)/(submitted reviews)"

    def __get_metrics(self):
        """ Each metric must have its own filters copy to modify it freely"""
        esfilters_merge = None
        esfilters_abandon = None
        esfilters_submit = None
        if self.esfilters:
            esfilters_merge = self.esfilters.copy()
            esfilters_abandon = self.esfilters.copy()
            esfilters_submit = self.esfilters.copy()

        merged = Merged(self.es_url, self.es_index,
                              start=self.start, end=self.end,
                              esfilters=esfilters_merge, interval=self.interval)
        # For BMI we need when the ticket was closed
        merged.DATE_FIELD = 'closed'

        abandoned = Abandoned(self.es_url, self.es_index,
                                    start=self.start, end=self.end,
                                    esfilters=esfilters_abandon, interval=self.interval)
        # For BMI we need when the ticket was closed
        abandoned.DATE_FIELD = 'closed'

        submitted = Submitted(self.es_url, self.es_index,
                                    start=self.start, end=self.end,
                                    esfilters=esfilters_submit,
                                    interval=self.interval)

        return (merged, abandoned, submitted)

    def get_agg(self):
        (merged, abandoned, submitted) = self.__get_metrics()
        merged_agg = merged.get_agg()
        abandoned_agg = abandoned.get_agg()
        closed_agg = merged_agg + abandoned_agg
        submitted_agg = submitted.get_agg()

        if submitted_agg == 0:
            bmi = 1  # if no submitted reviews, bmi is at 100%
        else:
            bmi = closed_agg/submitted_agg

        return bmi


    def get_ts(self):
        bmi = {}
        (merged, abandoned, submitted) = self.__get_metrics()
        merged_ts = merged.get_ts()
        abandoned_ts = abandoned.get_ts()
        submitted_ts = submitted.get_ts()

        bmi['date'] = merged_ts['date']
        bmi['unixtime'] = merged_ts['unixtime']
        bmi['value'] = []
        for i in range(0, len(submitted_ts['value'])):
            if submitted_ts['value'][i] == 0:
                bmi['value'].append(0)
            else:
                closed = merged_ts['value'][i] + abandoned_ts['value'][i]
                bmi['value'].append(closed / submitted_ts['value'][i])
        return bmi


class Organizations(GerritMetrics):
    id = "organizations"
    name = "Organizations"
    desc = "Number of organizations (organizations, etc.) with persons active in code review"
    FIELD_NAME = 'author_org_name'


class Projects(GerritMetrics):
    id = "projects"
    name = "Projects"
    desc = "Number of projects in code review"
    FIELD_NAME = 'project' # field used to list projects


class Submitters(GerritMetrics):
    id = "submitters"
    name = "Submitters"
    desc = "Number of persons submitting code review processes"
    FIELD_COUNT = 'author_uuid' # field used to count Authors
    FIELD_NAME = 'author_name' # field used to list Authors


class DaysToMergeMedian(GerritMetrics):
    id = "days_to_merge_review_median"
    name = "Days to merge reviews (median)"
    desc = "Number of days needed to merge a review (median)"
    FIELD_COUNT = 'timeopen'
    AGG_TYPE = 'median'
    filters = {"status": "MERGED"}

    def get_agg(self):
        agg = super(type(self), self).get_agg()
        if agg == None:
            agg = 0  # None is because NaN in ES. Let's convert to 0
        return agg

class DaysToMergeAverage(GerritMetrics):
    id = "days_to_merge_review_avg"
    name = "Days to merge reviews (average)"
    desc = "Number of days needed to merge a review (average)"
    FIELD_COUNT = 'timeopen'
    AGG_TYPE = 'average'
    filters = {"status": "MERGED"}

class PatchsetsMedian(GerritMetrics):
    id = "review_patchsets_median"
    name = "Patchsets per review (median)"
    desc = "Number of patchsets per review (median)"
    FIELD_COUNT = 'patchsets'
    AGG_TYPE = 'median'

    def get_agg(self):
        agg = super(type(self), self).get_agg()
        if agg == None:
            agg = 0  # None is because NaN in ES. Let's convert to 0
        return agg

class PatchsetsAverage(GerritMetrics):
    id = "review_patchsets_avg"
    name = "Patchsets per review (average)"
    desc = "Number of patchsets per review (average)"
    FIELD_COUNT = 'patchsets'
    AGG_TYPE = 'average'



#!/usr/bin/python3
## Copyright (C) 2016 Bitergia
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## This file is a part of GrimoireLib
##  (an Python library for the MetricsGrimoire and vizGrimoire systems)
##
##
## Authors:
##   Daniel Izquierdo-Cortazar <dizquierdo@bitergia.com>
##   Alvaro del Castillo <acs@bitergia.com>

import logging

import pytz

import requests

from datetime import datetime, timedelta

from ..esquery import ElasticQuery


logger = logging.getLogger(__name__)


class Metrics(object):
    """Root of hierarchy of Entities (Metrics)

    This is the root hierarchy of the of the metric classes
    defined for each of the data sources.

    When instantiated Metrics, we can obtain specific representations
    of such entity. Timeseries of datasets (get_ts method), aggregated
    data (get_agg method) or lists of elements (get_list method).

    """

    id = None
    name = None
    desc = None
    AGG_TYPE = 'count'  # agg type for the metric
    FIELD_NAME = None  # Field used for metric lists
    FIELD_COUNT = None  # Field used for metric count
    FIELD_DATE = 'grimoire_creation_date'
    DEFAULT_INTERVAL = '1M'
    filters = None  # fixed filters for the metric
    filters_core = None  # Core filters to be used in all metrics
    interval = '1M'  # interval to be used in all metrics
    offset = None  # offset to be used in date histogram in all metrics

    def __init__(self, es_url, es_index, start=None, end=None, esfilters={},
                 interval=None, offset=None):
        """es connection and filter to be used"""
        self.es_url = es_url
        self.es_index = es_index
        self.start = start
        self.end = end
        self.esfilters = esfilters if esfilters else {}
        if self.filters:
            # If there are metric class filters use them also
            self.esfilters.update(self.filters)
        if self.filters_core:
            # If there are core filters for all metrics use them also
            self.esfilters.update(self.filters_core)
        if interval:
            self.interval = interval
        if offset:
            self.offset = offset

    def get_definition(self):
        def_ = {
               "id":self.id,
               "name":self.name,
               "desc":self.desc
        }
        return def_

    def get_query(self, evolutionary=False):
        """
        Private method that returns a valid ElasticSearch query for
        computing the metric.

        :evolutionary: boolean
            If True, an evolutionary analysis is provided
            If False, an aggregated analysis is provided
        """

        if not evolutionary:
            interval = None
            offset = None
        else:
            interval = self.interval
            offset = self.offset
            if not interval:
                raise RuntimeError("Evolutionary query without an interval.")

        query = ElasticQuery.get_agg(field=self.FIELD_COUNT,
                                     date_field=self.FIELD_DATE,
                                     start=self.start, end=self.end,
                                     filters=self.esfilters,
                                     agg_type=self.AGG_TYPE,
                                     interval=interval,
                                     offset=offset)

        logger.debug("Metric: '%s' (%s); Query: %s",
                     self.name, self.id, query)
        return query


    def get_list(self):
        field = self.FIELD_NAME
        query = ElasticQuery.get_agg(field=field,
                                     date_field=self.FIELD_DATE,
                                     start=self.start, end=self.end,
                                     filters=self.esfilters)
        logger.debug("Metric: '%s' (%s); Query: %s",
                     self.name, self.id, query)
        res = self.get_metrics_data(query)
        l = {field:[],"value":[]}
        for bucket in res['aggregations'][str(ElasticQuery.AGGREGATION_ID)]['buckets']:
            l[field].append(bucket['key'])
            l['value'].append(bucket['doc_count'])
        return l

    def get_metrics_data(self, query):
        """ Get the metrics data from ES """
        url = self.es_url+'/' + self.es_index + '/_search'
        r = requests.post(url, query)
        r.raise_for_status()
        return r.json()

    def get_ts(self):
        """Returns a time series of a specific class

        A timeseries consists of a unixtime date, labels, some other
        fields and the data of the specific instantiated class per
        interval. This is built on a hash table.

        """
        query = self.get_query(True)
        res = self.get_metrics_data(query)
        # Time to convert it to our grimoire timeseries format
        ts = {"date":[],"value":[],"unixtime":[]}
        agg_id = ElasticQuery.AGGREGATION_ID
        if 'buckets' not in res['aggregations'][str(agg_id)]:
            raise RuntimeError("Aggregation results have no buckets in time series results.")
        for bucket in res['aggregations'][str(agg_id)]['buckets']:
            ts['date'].append(bucket['key_as_string'])
            if str(agg_id+1) in bucket:
                # We have a subaggregation with the value
                # If it is percentiles we get the median
                if 'values' in bucket[str(agg_id+1)]:
                    val = bucket[str(agg_id+1)]['values']['50.0']
                    if val == 'NaN':
                        # ES returns NaN. Convert to None for matplotlib graph
                        val = None
                    ts['value'].append(val)
                else:
                    ts['value'].append(bucket[str(agg_id+1)]['value'])
            else:
                ts['value'].append(bucket['doc_count'])
            # unixtime comes in ms from ElasticSearch
            ts['unixtime'].append(bucket['key']/1000)
        return ts

    def get_agg(self):
        """ Returns an aggregated value """
        query = self.get_query(False)
        res = self.get_metrics_data(query)
        # We need to extract the data from the JSON res
        # If we have agg data use it
        agg_id = str(ElasticQuery.AGGREGATION_ID)
        if 'aggregations' in res and 'values' in res['aggregations'][agg_id]:
            if self.AGG_TYPE=='median':
                agg = res['aggregations'][agg_id]['values']["50.0"]
                if agg == 'NaN':
                    # ES returns NaN. Convert to None for matplotlib graph
                    agg = None
            else:
                raise RuntimeError("Multivalue aggregation result not supported")
        elif 'aggregations' in res and 'value' in res['aggregations'][agg_id]:
            agg = res['aggregations'][agg_id]['value']
        else:
            agg = res['hits']['total']

        return agg


    def get_trend(self):
        """ Get the trends for the interval defined in the metric """

        # TODO: We just need the last two periods, not the full ts
        ts = self.get_ts()
        last = ts['value'][len(ts['value'])-1]
        prev = ts['value'][len(ts['value'])-2]

        trend = last - prev
        trend_percentage = None
        if last == 0:
            if prev > 0:
                trend_percentage = -100
            else:
                trend_percentage = 0
        else:
            trend_percentage = int((trend/last)*100)

        return (last, trend_percentage)
